/***************************************************************************

					Generate Z80 CPU Opcodes

	Doesn't generate

	(Base Set)			76 HALT			F3 DI			FB EI
	(Extend CB)			(None)
	(Extend ED)			NEG RETN IMn RETI IM2 RLD RRD and move commands

 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

char cSet = '_';						/* _ for norm C for xCB E for xED */

int bCFlag = 1;

#define DEF(op)		DefineOpcode('_',op)
#define DEFI(op,ix)	DefineOpcode(ix,op)

void DoNormal(void);
void Section(char *szName);
void DefineOpcode(char cGroup,int nOpcode);
void Generate1(void);
void Generate2(void);
void Generate3(void);
void Generate4(void);
void Generate5(void);
void Generate6(void);
void Generate7(void);
void Generate8(void);
void Generate9(void);
void Generate10(void);
void DoExtendCB(void);
void DoExtendED(void);
void xRead(int nReg);
void xWrite(int nReg);

void GenReturn(char *szCond);
void RGet(char *szPC,int nReg);
void RPut(char *szPC,int nReg);
void RPtr(int nReg);

#define R_B		(0)						/* Register IDs */
#define R_C		(1)
#define R_D		(2)
#define R_E		(3)
#define R_H		(4)
#define R_L		(5)
#define R_HL	(6)
#define R_A		(7)
#define R_IXn	(8)
#define R_IYn	(9)

static int nCount = 0;
static int nCCount = 0;
static int nECount = 0;

char *szReg[] = { "b","c","d","e","h","l","(hl)","<a reg>","(ix+n)","(iy+n)" };

#define REG(n)	(szReg[n])				/* Get register name */

void Clock(n)
{
/*  if (bCFlag == 1) exit(0); */
bCFlag = 1;
printf(" clock %d\n",n);
}

	/*****************************/
	/* Generate a section header */
	/*****************************/

void Section(char *szName)
{
printf("; =========================================================================\n");
printf(";\n;%*s%s\n;\n",40-strlen(szName)/2,"",szName);
printf("; =========================================================================\n");
}

	/***********************/
	/* Define a new opcode */
	/***********************/

static int bFirst = 1;

void DefineOpcode(char cGroup,int nOpcode)
{
/* if (bCFlag == 0) exit(0);       */
bCFlag = 0;
if (!bFirst) printf(" ret\n");			/* Close the last one automatically */
bFirst = 0;
if (nOpcode >= 0)
	{
	if (cSet == '_' && cGroup == '_') nCount++;
	if (cSet == 'C' && cGroup == '_') nCCount++;
	if (cSet == 'E' && cGroup == '_') nCCount++;
	printf("; =========================================================================\n");
	printf("_%c%c%02x:\n",cGroup,cSet,nOpcode);
	}
}

	/*****************************************************************/
	/* Do the normal opcodes - the default set, and their duplicates */
	/* used in IX and IY											 */
	/*****************************************************************/

void DoNormal(void)
{
Generate1();
Generate2();
Generate3();
Generate4();
Generate5();
Generate6();
Generate7();
Generate8();
Generate9();
Generate10();
}

	/************************************************************/
	/* Generate the register to register moves where a register */
	/* is b,c,d,e,h,l,(hl),a,(ix+n),(iy+n)						*/
	/************************************************************/

void Generate1()
{
int nSrc,nTgt,nOpcode,n,x;
Section("Register -> Register (b,c,d,e,h,l,(hl),a,(ix+n),(iy+n)");
for (nSrc = R_B;nSrc <= R_A;nSrc++)
	for (nTgt = R_B;nTgt <= R_A;nTgt++)
		{
		nOpcode = 0x40 + nTgt * 8 + nSrc;
		if (nOpcode != 0x76)
			{
			DEF(nOpcode);
			Clock((nSrc == R_HL || nTgt == R_HL) ? 7 : 4);
			if (nSrc != nTgt)
				{
				RGet("al",nSrc);
				RPut("al",nTgt);
				}
			}
		}
for (n = R_B;n <= R_A;n++)
	{
	if (n != R_HL)
		{
		for (x = R_IXn;x <= R_IYn;x++)
			{
			DEFI(0x70+n,(x == R_IXn) ? 'X' : 'Y');
			Clock(19);
			RGet("al",n);
			RPut("al",x);
			DEFI(0x46+n*8,(x == R_IXn) ? 'X' : 'Y');
			Clock(19);
			RGet("al",x);
			RPut("al",n);
			}
		}
	}
}

	/**************************************************************/
	/* Generate the <operation> a,<register> commands. Operations */
	/* are add adc sub sbc and xor or cp registers as above		  */
	/**************************************************************/

char *szOp[] = { "add","adc","sub","sbc","and","xor","or","cp" };

void Generate2(void)
{
int nOp,nReg,nOpcode;
Section("Register to Register arithmetic operations");
for (nOp = 0;nOp < 8;nOp++)
	{
	for (nReg = R_B;nReg <= R_IYn;nReg++)
		{
		nOpcode = (0x80 + nOp * 8 + nReg);
		if (nReg == R_IXn || nReg == R_IYn)
			{
			DEFI(0x86+nOp*8,(nReg == R_IXn ? 'X' : 'Y'));
			Clock(19);
			}
		else
			{
			DEF(nOpcode);
			Clock((nReg == R_HL) ? 7 : 4);
			}
		RGet("ah",nReg);
		RGet("al",R_A);
		printf(" alu_%s\n",szOp[nOp]);
		if (nOp != 7) RPut("al",R_A);	/* Cmp doesn't update accumulator */
		}
	DEF(0xC6+nOp*8);
	printf(" clock 7\n");
	printf(" fetch ah\n");
	RGet("al",R_A);
	printf(" alu_%s\n",szOp[nOp]);
	if (nOp != 7) RPut("al",R_A);	/* Cmp doesn't update accumulator */
	}
}

	/*****************************************/
	/* Generate inc,dec,ld for the registers */
	/*****************************************/

void Generate3()
{
int nReg,nOpBase,cSection,bIX;

Section("Inc/Dec/Load registers");
for (nReg = R_B;nReg <= R_IYn;nReg++)
	{
	cSection = '_';
	nOpBase = 0x00 + nReg * 8;
	bIX = 0;
	if (nReg == R_IXn || nReg == R_IYn)
		{
		cSection = (nReg == R_IXn) ? 'X' : 'Y';
		nOpBase = 0x30;
		bIX = 1;
		}
	if (cSection == '_' && nReg != 6)
		{
		DEF(nOpBase+4);
		Clock(4);
		RGet("al",nReg);
		printf(" alu_inc8\n");
		RPut("al",nReg);
		DEF(nOpBase+5);
		Clock(4);
		RGet("al",nReg);
		printf(" alu_dec8\n");
		RPut("al",nReg);
		DEF(nOpBase+6);
		printf(" fetch al\n");
		RPut("al",nReg);
		Clock((nReg == 6) ? 7 : 4);
		}
	else
		{
		DEFI(nOpBase+4,cSection);
		Clock(bIX ? 23 : 11);
		RPtr(nReg);
		printf(" push bx\n");
		printf(" read al\n");
		printf(" alu_inc8\n");
		printf(" pop bx\n");
		printf(" write al\n");

		DEFI(nOpBase+5,cSection);
		Clock(bIX ? 23 : 11);
		RPtr(nReg);
		printf(" push bx\n");
		printf(" read al\n");
		printf(" alu_dec8\n");
		printf(" pop bx\n");
		printf(" write al\n");

		DEFI(nOpBase+6,cSection);
		RPtr(nReg);
		Clock(bIX ? 19 :10);
		printf(" push bx\n");
		printf(" fetch al\n");
		printf(" pop bx\n");
		printf(" write al\n");
		}
	}
}

	/****************************************/
	/* Does the double register command set */
	/****************************************/

char *szDReg[] = { "bc","de","hl","sp","ix","iy" };

void Generate4(void)
{
int nReg,nOpBase,n1,n2,n3;
char c,*pHl;
Section("Double registers");
for (nReg = 0;nReg < 6;nReg++)
	{
	nOpBase = nReg * 0x10;
	c = '_';
	if (nReg >= 4)
		{
		nOpBase = 0x20;
		c = (nReg == 4 ? 'x' : 'y');
		}
	DEFI(nOpBase + 1,c);
	printf(" fetchw\n");
	printf(" mov [_%s],ax\n",szDReg[nReg]);
	Clock(10);
	DEFI(nOpBase + 2,c);
	switch(nReg)
		{
		case 0:
		case 1:	printf(" mov bx,[_%s]\n",szDReg[nReg]);
				RGet("al",R_A);
				printf(" write al\n");
				Clock(7);
				break;
		case 4:
		case 5:
		case 2:	printf(" fetchw\n");
				printf(" mov bx,ax\n");
				printf(" mov ax,[_%s]\n",szDReg[nReg]);
				printf(" writew\n");
				Clock(20);
				break;
		case 3:	printf(" fetchw\n");
				printf(" mov bx,ax\n");
				RGet("al",R_A);
				printf(" write al\n");
				Clock(13);
				break;
		}
	DEFI(nOpBase+3,c);
	Clock(nReg >= 4 ? 10:6);
	printf(" mov ax,[_%s]\n",szDReg[nReg]);
	printf(" alu_inc16\n");
	printf(" mov [_%s],ax\n",szDReg[nReg]);

	pHl = "hl";
	if (nReg == 4) pHl = "ix";
	if (nReg == 5) pHl = "iy";

	n3 = nReg;

	DEFI(nOpBase+9,c);
	printf(" mov ax,[_%s]\n",pHl);
	printf(" mov bx,[_%s]\n",szDReg[n3]);
	printf(" alu_add16\n");
	printf(" mov [_%s],ax\n",pHl);
	Clock(11);
	DEFI(nOpBase + 10,c);
	switch(nReg)
		{
		case 0:
		case 1:	printf(" mov bx,[_%s]\n",szDReg[nReg]);
				printf(" read al\n");
				RPut("al",R_A);
				Clock(7);
				break;
		case 4:
		case 5:
		case 2:	printf(" fetchw\n");
				printf(" mov bx,ax\n");
				printf(" readw\n");
				printf(" mov [_%s],ax\n",szDReg[nReg]);
				Clock(20);
				break;
		case 3:	printf(" fetchw\n");
				printf(" mov bx,ax\n");
				printf(" read al\n");
				RPut("al",R_A);
				Clock(13);
				break;
		}
	DEFI(nOpBase+11,c);
	Clock(nReg >= 4 ? 10:6);
	printf(" mov ax,[_%s]\n",szDReg[nReg]);
	printf(" alu_dec16\n");
	printf(" mov [_%s],ax\n",szDReg[nReg]);
	}
for (n1 = 'x';n1 <= 'y';n1++)
	{
	for (nReg = 0;nReg < 4;nReg++)
		{
		if (nReg != 2)
			{
			DEFI(0x9+nReg*0x10,n1);
			printf(" mov ax,[_I%c]\n",n1);
			printf(" mov bx,[_%s]\n","bc\0de\0xx\0sp"+nReg*3);
			printf(" alu_add16\n");
			printf(" mov [_I%c],ax\n",n1);
			Clock(15);
			}
		}
	}
}

	/******************/
	/* Relative jumps */
	/******************/

char *szCond[] = { "nz","z","nc","c","np","p","ns","s" };

void Generate5(void)
{
int i;
Section("Relative Jumps");
for (i = 0;i < 5;i++)
	{
	DEF(i == 4 ? 0x18 : 0x20 + i * 8);
	printf(" eac_relative\n");
	Clock(7);
	if (i < 4) GenReturn(szCond[i]);
	Clock(5);
	printf(" add bp,bx\n");
	}
}

	/**************************/
	/* Generate Calls & Jumps */
	/**************************/

void Generate6(void)
{
int n;
Section("Returns, Calls and Jumps");
for (n = 0;n < 8;n++)
	{
	DEF(n * 8 + 0xC0);
	Clock(5);
	GenReturn(szCond[n]);
	printf(" rpull\n");
	printf(" mov bp,ax\n");
	Clock(6);

	DEF(n * 8 + 0xC2);
	Clock(10);
	printf(" fetchw\n");
	GenReturn(szCond[n]);
	printf(" mov bp,ax\n");

	DEF(n * 8 + 0xC4);
	Clock(7);
	printf(" fetchw\n");
	GenReturn(szCond[n]);
	Clock(10);
	printf(" push ax\n");
	printf(" mov ax,bp\n");
	printf(" rpush\n");
	printf(" pop bp\n");
	}

DEF(0xC9);
Clock(10);
printf(" rpull\n");
printf(" mov bp,ax\n");

DEF(0xC3);
Clock(10);
printf(" fetchw\n");
printf(" mov bp,ax\n");

DEF(0xCD);
Clock(17);
printf(" fetchw\n");
printf(" push ax\n");
printf(" mov ax,bp\n");
printf(" rpush\n");
printf(" pop bp\n");
}

	/*********************/
	/* Generate the RSTs */
	/*********************/

void Generate7()
{
int i;
Section("Rst xxx");
for (i = 0;i < 8;i++)
	{
	DEF(i*8+0xC7);
	Clock(11);
	printf(" mov ax,bp\n");
	printf(" rpush\n");
	printf(" mov bp,%d\n",i*8);
	}
}

	/********************************/
	/* Generate Rxx rotate commands */
	/********************************/

char *szRotates[] = { "rlc","rrc","rl","rr","sla","sra",NULL,"srl" };

void Generate8(void)
{
int i;
Section("Rotate/Shift Accumulator");
for (i = 0;i < 4;i++)
	{
	DEF(i*8+7);
	Clock(4);
	RGet("al",R_A);
	printf(" alu_%s\n",szRotates[i]);
	RPut("al",R_A);
	}
}

	/********************/
	/* Pulls and pushes */
	/********************/

char *szDPR[] = { "bc","de","hl","_af","ix","iy" };

void Generate9(void)
{
int nReg,nBase;
char cSection;
char *p;
Section("Push and Pull\n");
for (nReg = 0;nReg < 6;nReg++)
	{
	p = szDPR[nReg];
	nBase = 0xC1 + nReg * 16;
	cSection = '_';
	if (nReg > 3)
		{
		nBase = 0xE1;
		cSection = (nReg == 4) ? 'x':'y';
		}

	DEFI(nBase,cSection);
	Clock((nReg > 3) ? 15 : 11);
	printf(" rpull\n");
	if (nReg == 3)
		printf(" mov cx,ax\n");
	else
		printf(" mov [_%s],ax\n",p);
	if (nReg == 3) printf(" call UpdateByteRegisters\n");

	DEFI(nBase+4,cSection);
	Clock((nReg > 3) ? 15 : 11);
	if (nReg == 3) printf(" call UpdateFlagBits\n");
	if (nReg == 3)
		printf(" mov ax,cx\n");
	else
		printf(" mov ax,[_%s]\n",p);
	printf(" rpush\n");
	}
}

	/***********************/
	/* Miscellaneous stuff */
	/***********************/

void Generate10(void)
{
int n;
char *s,c;

Section("Miscellaneous");
DEF(00);							/* No operation */
Clock(4);

DEF(0x8);							/* EX af 'af */
printf(" call UpdateFlagBits\n");
printf(" mov bx,[_xaf]\n");
printf(" mov [_xaf],cx\n");
printf(" mov cx,bx\n");
printf(" call UpdateByteRegisters\n");
Clock(4);

DEF(0x10);							/* DJNZ B */
printf(" mov al,[_b]\n");
printf(" alu_dec8\n");
printf(" mov [_b],al\n");
Clock(8);
printf(" eac_relative\n");
GenReturn("nz");
Clock(5);
printf(" add bp,bx\n");

DEF(0x27);							/* DAA */
Clock(4);
printf(" alu_daa\n");

DEF(0x2F);							/* CPL */
Clock(4);
RGet("al",R_A);
printf(" xor al,255\n");
printf(" setf NFlag+HFlag\n");
RPut("al",R_A);

DEF(0x37);							/* SCF */
Clock(4);
printf(" setf CFlag\n");

DEF(0x3F);							/* CCF */
Clock(4);
printf(" xor cl,CFlag\n");

DEF(0xD3);							/* Out n,a */
Clock(11);
printf(" fetch ah\n");
RGet("al",R_A);
printf(" cpu_out\n");

DEF(0xD9);							/* EXX */
Clock(4);
printf(" wswap _bc,_xbc\n");
printf(" wswap _de,_xde\n");
printf(" wswap _hl,_xhl\n");

DEF(0xDB);							/* In a,n */
Clock(11);
printf(" fetch ah\n");
printf(" cpu_in\n");
RPut("al",R_A);

DEF(0xEB);
Clock(4);
printf(" wswap _de,_hl\n");		/* Ex de,hl */


for (n = 0;n < 3;n++)
	{
	c = ("_xy")[n];
	s = ("hl\0ix\0iy\0")+n*3;
	DEFI(0xE3,c);
	Clock((n > 0) ? 23 : 4);
	printf(" mov bx,[_sp]\n");
	printf(" readw\n");
	printf(" push ax\n");
	printf(" mov bx,[_sp]\n");
	printf(" mov ax,[_%s]\n",s);
	printf(" writew\n");
	printf(" pop ax\n");
	printf(" mov [_%s],ax\n",s);

	DEFI(0xF9,c);
	Clock((n > 0) ? 10 : 6);
	printf(" mov ax,[_%s]\n",s);
	printf(" mov [_sp],ax\n");

	DEFI(0xE9,c);
	Clock((n > 0) ? 8 : 4);
	printf(" mov bp,[_%s]\n",s);
	}

}
	/********************************************/
	/* Generate return if given condition FAILS */
	/********************************************/

void GenReturn(char *szCond)
{
int nRSet;
char cFlag = szCond[strlen(szCond)-1];
cFlag = toupper(cFlag);
if (cFlag == 'Z' || cFlag == 'S')
				printf(" call UpdateFlagBits\n");
printf(" testf %cFlag\n",cFlag);
nRSet = (szCond[0] == 'n');
printf(" j%s ret\n",nRSet ? "nz" : "z");
}

	/*****************************************************************/
	/* Access registers. RGet puts register into szPC reg, RPut puts */
	/* szPC reg into Z80 reg, RPtr puts pointer to reg into BX		 */
	/*****************************************************************/

void RGet(char *szPC,int nReg)
{
if (nReg == R_IXn || nReg == R_IYn || nReg == R_HL)
	{
	RPtr(nReg);
	printf(" read %s\n",szPC);
	}
else
	{
	if (nReg == R_A)
		printf(" mov %s,ch\n",szPC);
	else
		printf(" mov %s,[_%s]\n",szPC,REG(nReg));
	}
}

void RPut(char *szPC,int nReg)
{
if (nReg == R_IXn || nReg == R_IYn || nReg == R_HL)
	{
	RPtr(nReg);
	printf(" write %s\n",szPC);
	}
else
	{
	if (nReg == R_A)
		printf(" mov ch,%s\n",szPC);
	else
		printf(" mov [_%s],%s\n",REG(nReg),szPC);
	}
}

void RPtr(int nReg)
{
switch(nReg)
	{
	case	R_IXn:
	case	R_IYn:	printf(" eac_ixoffset\n");
					printf(" add bx,[_I%c]\n",(nReg == R_IXn) ? 'X' : 'Y');
					break;
	case	R_HL:	printf(" mov bx,[_HL]\n");
					break;
	default:		exit(fprintf(stderr,"Tried to RPtr %s\n",REG(nReg)));
					break;
	}
}

	/*******************************/
	/* Section (extended CB Codes) */
	/*******************************/

void DoExtendCB(void)
{
int nReg,nShift,nBit,nMask;
cSet = 'C';
for (nReg = R_B;nReg <= R_A;nReg++)
	{
	for (nShift = 0;nShift < 8;nShift++)
		{
		if (szRotates[nShift] != NULL)
			{
			DEF(0x00 + nReg + nShift * 8);
			Clock(nReg == R_HL ? 15 : 8);
			xRead(nReg);
			printf(" alu_%s\n",szRotates[nShift]);
			printf(" setzn al\n");
			printf(" setparity\n");
			xWrite(nReg);
			}
		}
	for (nBit = 0;nBit < 8;nBit++)
		{
		DEF(0xC0 + nBit * 8 + nReg);	/* SET */
		Clock(nReg == R_HL ? 15 : 8);
		nMask = (1 << nBit);
		switch(nReg)
			{
			case 6:
				printf(" mov bx,[_eac]\n");
				printf(" read al\n");
				printf(" or al,%d\n",nMask);
				printf(" mov bx,[_eac]\n");
				printf(" write al\n");
				break;
			case 7:
				printf(" or ch,%d\n",nMask);
				break;
			default:
				printf(" or byte [_%s],%d\n",REG(nReg),nMask);
				break;
			}

		DEF(0x80 + nBit * 8 + nReg);	/* RESET */
		Clock(nReg == R_HL ? 15 : 8);
		switch(nReg)
			{
			case 6:
				printf(" mov bx,[_eac]\n");
				printf(" read al\n");
				printf(" and al,%d\n",255-nMask);
				printf(" mov bx,[_eac]\n");
				printf(" write al\n");
				break;
			case 7:
				printf(" and ch,%d\n",255-nMask);
				break;
			default:
				printf(" and byte [_%s],%d\n",REG(nReg),255-nMask);
				break;
			}

		DEF(0x40 + nBit * 8 + nReg);	/* BIT */
		Clock(nReg == R_HL ? 12 : 8);
		RGet("al",nReg);
		printf(" mov ah,%d\n",nMask);
		printf(" alu_bit\n");
		}
	}
}

	/*****************************************************************/
	/* These read and write registers for the CB set. The difference */
	/* is that if the register is HL it uses the _EAC set before     */
	/* the _CB code is entered										 */
	/*****************************************************************/

void xRead(int nReg)
{
if (nReg == R_HL)
	{
	printf(" mov bx,[_EAC]\n");
	printf(" read al\n");
	}
else
	{
	if (nReg == R_A)
		printf(" mov al,ch\n");
	else
		printf(" mov al,[_%s]\n",REG(nReg));
	}
}

void xWrite(int nReg)
{
if (nReg == R_HL)
	{
	printf(" mov bx,[_EAC]\n");
	printf(" write al\n");
	}
else
	{
	if (nReg == R_A)
		printf(" mov ch,al\n");
	else
		printf(" mov [_%s],al\n",REG(nReg));
	}
}

	/****************************************/
	/* Do some of the extended (ED) opcodes */
	/****************************************/

char *szXFr[] = { "ldi","cpi","ini","outi","ldd","cpd","ind","outd" };

void DoExtendED(void)
{
int i,nReg;
char *p;
cSet = 'E';
for (nReg = R_B;nReg <= R_A;nReg++)
	{
	if (nReg != R_HL)
		{
		DEF(0x40+nReg * 8);
		Clock(12);
		RGet("ah",R_C);
		printf(" cpu_in\n");
		RPut("al",nReg);
		printf(" setzn al\n");
		DEF(0x41+nReg * 8);
		Clock(12);
		RGet("al",nReg);
		RGet("ah",R_C);
		printf(" cpu_out\n");
		}
	}
for (nReg = 0;nReg < 4;nReg++)
	{
	p = "bc\0de\0hl\0sp"+nReg*3;
	DEF(0x42+nReg * 16);
	printf(" mov ax,[_hl]\n");
	printf(" mov bx,[_%s]\n",p);
	printf(" alu_sbc16\n");
	printf(" mov [_hl],ax\n");
	Clock(15);

	DEF(0x43+nReg*16);
	printf(" fetchw\n");
	printf(" mov bx,ax\n");
	printf(" mov ax,[_%s]\n",p);
	printf(" writew\n");
	Clock(20);

	DEF(0x4A+nReg * 16);
	printf(" mov ax,[_hl]\n");
	printf(" mov bx,[_%s]\n",p);
	printf(" alu_adc16\n");
	printf(" mov [_hl],ax\n");
	Clock(15);

	DEF(0x4B+nReg*16);
	printf(" fetchw\n");
	printf(" mov bx,ax\n");
	printf(" readw\n");
	printf(" mov [_%s],ax\n",p);
	Clock(20);
	}

for (i = 0;i < 8;i++)
	{
	DEF(0xA0 + ((i >= 4) ? i + 4 : i));
	printf(" block_%s\n",szXFr[i]);
	Clock(21);

	DEF(0xB0 + ((i >= 4) ? i + 4 : i));
	printf("_block%d:\n",i);
	printf(" block_%s\n",szXFr[i]);
	Clock(21+16);

	switch(i % 4)
		{
		case	0:	printf(" cmp word [_bc],0\n");
					printf(" jz ret\n");
					break;

		case	1:	printf(" call UpdateFlagBits\n");
					printf(" testf ZFlag\n");
					printf(" jnz ret\n");
					printf(" cmp word [_bc],0\n");
					printf(" jz ret\n");
					break;
		case	2:
		case	3:	printf(" cmp byte [_b],0\n");
					printf(" jnz ret\n");
		}
	printf(" jmp _block%d\n",i);
	}
}

void main()
{
int i;
Section("This code is generated automatically. Do not edit it !!!!!");
printf("\n\n\n");


/* Do the base codes, and the base codes extended by FD or DD */
DoNormal();

/* Do the CB Codes */
DoExtendCB();

/* Do the ED codes */
DoExtendED();

DEF(-1);								/* Close the last one */
printf(";\n; Generated %d in the base set.\n;\n",nCount);
printf(";\n; Generated %d in the xCB set.\n;\n",nCCount);
printf(";\n; Generated %d in the xED set.\n;\n",nECount);

}